"""App Center analytics API wrappers."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import enum
import logging
from typing import List, Union

import deserialize

from appcenter.derived_client import AppCenterDerivedClient
from appcenter.models import UserToken


class AppCenterTokensClient(AppCenterDerivedClient):
    """Wrapper around the App Center tokens APIs.

    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    class TokenScope(enum.Enum):
        """Specifies the scope of a token."""

        FULL = "all"
        READER = "viewer"

    def __init__(self, token: str, parent_logger: logging.Logger) -> None:
        super().__init__("tokens", token, parent_logger)

    def get_user_tokens(self) -> List[UserToken]:
        """Get the users tokens

        :returns: The tokens
        """

        self.log.info("Getting user tokens")

        request_url = self.base_url()
        request_url += "/api_tokens"

        response = self.get(request_url)

        return deserialize.deserialize(List[UserToken], response.json())

    def create_user_token(self, name: str, scope: TokenScope) -> UserToken:
        """Create a user token.

        :param name: The name/description of the token
        :param scope: The scope for the token
        """

        request_url = self.base_url()
        request_url += "/api_tokens"

        self.log.debug(f"Creating user token name={name}, scope={scope}")

        response = self.post(
            request_url, data={"description": name, "scope": [scope.value]}
        )

        return deserialize.deserialize(UserToken, response.json())

    def delete_user_token(self, token: Union[str, UserToken]) -> None:
        """Delete a user token.

        :param token: The token itself or the ID for the token
        """

        request_url = self.base_url()
        request_url += "/api_tokens/"

        if isinstance(token, str):
            self.log.debug(f"Deleting user token={token}")
            request_url += token
        else:
            self.log.debug(f"Deleting user token={token.identifier}")
            request_url += token.identifier

        _ = self.delete(request_url)



"""App Center versions API wrappers."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import os
import time
from typing import Iterator, List, Optional
import urllib.parse

import deserialize

from appcenter.derived_client import AppCenterDerivedClient
from appcenter.models import (
    BasicReleaseDetailsResponse,
    BuildInfo,
    ChunkUploadResponse,
    ReleaseDetailsResponse,
    CreateReleaseUploadResponse,
    CommitUploadResponse,
    ReleaseDestinationResponse,
    ReleaseUpdateRequest,
    SetUploadMetadataResponse,
    UploadCompleteResponse,
)


# From https://github.com/microsoft/appcenter-cli/blob/master/appcenter-file-upload-client-node/src/ac-fus-mime-types.ts
MIME_TYPES = {
    "apk": "application/vnd.android.package-archive",
    "aab": "application/vnd.android.package-archive",
    "msi": "application/x-msi",
    "plist": "application/xml",
    "aetx": "application/c-x509-ca-cert",
    "cer": "application/pkix-cert",
    "xap": "application/x-silverlight-app",
    "appx": "application/x-appx",
    "appxbundle": "application/x-appxbundle",
    "appxupload": "application/x-appxupload",
    "appxsym": "application/x-appxupload",
    "msix": "application/x-msix",
    "msixbundle": "application/x-msixbundle",
    "msixupload": "application/x-msixupload",
    "msixsym": "application/x-msixupload",
}


class AppCenterVersionsClient(AppCenterDerivedClient):
    """Wrapper around the App Center versions APIs.

    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    def __init__(self, token: str, parent_logger: logging.Logger) -> None:
        super().__init__("versions", token, parent_logger)

    def recent(
        self, *, owner_name: str, app_name: str
    ) -> List[BasicReleaseDetailsResponse]:
        """Get the recent version for each distribution group.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app

        :returns: A list of BasicReleaseDetailsResponse
        """

        self.log.info(f"Getting recent versions of app: {owner_name}/{app_name}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/recent_releases"

        response = self.get(request_url)

        return deserialize.deserialize(
            List[BasicReleaseDetailsResponse], response.json()
        )

    def all(
        self,
        *,
        owner_name: str,
        app_name: str,
        published_only: bool = False,
        scope: Optional[str] = None,
    ) -> Iterator[BasicReleaseDetailsResponse]:
        """Get all (the 100 latest) versions.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param bool published_only: Return only the published releases (defaults to false)
        :param Optional[str] scope: When the scope is 'tester', only includes
                                    releases that have been distributed to
                                    groups that the user belongs to.

        :returns: An iterator of BasicReleaseDetailsResponse
        """

        self.log.info(f"Getting versions of app: {owner_name}/{app_name}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/releases?"

        parameters = {"published_only": str(published_only).lower()}

        if scope:
            parameters["scope"] = scope

        request_url += urllib.parse.urlencode(parameters)

        response = self.get(request_url)

        return deserialize.deserialize(
            List[BasicReleaseDetailsResponse], response.json()
        )

    def release_details(
        self, *, owner_name: str, app_name: str, release_id: int
    ) -> ReleaseDetailsResponse:
        """Get the full release details for a given version.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param int release_id: The ID of the release to get the details for

        :returns: The full details for a release
        """

        self.log.info(f"Getting details for: {owner_name}/{app_name} - {release_id}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/releases/{release_id}?"

        response = self.get(request_url)

        return deserialize.deserialize(ReleaseDetailsResponse, response.json())

    def release_id_for_version(
        self, *, owner_name: str, app_name: str, version: str
    ) -> Optional[int]:
        """Get the App Center release identifier for the app version (usually build number).

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param bool version: The app version (usually build number)

        :returns: The App Center release identifier
        """

        for app_version in self.all(owner_name=owner_name, app_name=app_name):
            if app_version.version == version:
                return app_version.identifier

        return None

    def latest_commit(self, *, owner_name: str, app_name: str) -> Optional[str]:
        """Find the most recent release which has an available commit in it and return the commit hash.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app

        :returns: The latest commit available on App Center
        """

        self.log.info(f"Getting latest commit for app: {owner_name}/{app_name}")

        for version in self.all(owner_name=owner_name, app_name=app_name):
            full_details = self.release_details(
                owner_name=owner_name, app_name=app_name, release_id=version.identifier
            )

            if full_details.build is None:
                continue

            if full_details.build.commit_hash is not None:
                return full_details.build.commit_hash

        return None

    def get_upload_url(
        self, *, owner_name: str, app_name: str
    ) -> CreateReleaseUploadResponse:
        """Get the App Center release identifier for the app version (usually build number).

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app

        :returns: The App Center release identifier
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/uploads/releases"

        for attempt in range(3):
            self.log.debug(f"Attempting post {attempt}/3 in get_upload_url")
            try:
                response = self.post(request_url, data={})
                if response.ok:
                    break
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to post in get_upload_url: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return deserialize.deserialize(CreateReleaseUploadResponse, response.json())

    def set_upload_metadata(
        self,
        *,
        create_release_upload_response: CreateReleaseUploadResponse,
        binary_path: str,
    ) -> Optional[SetUploadMetadataResponse]:
        """Set the metadata for a binary upload

        :param CreateReleaseUploadResponse create_release_upload_response: The response to a `get_upload_url` call
        :param binary_path: The path to the binary to upload

        :returns: The upload response if we manage to get one, None otherwise
        """
        file_size = os.path.getsize(binary_path)
        file_name = os.path.basename(binary_path)
        file_ext = os.path.splitext(file_name)[-1]
        if file_ext.startswith("."):
            file_ext = file_ext[1:]
        mime_type = MIME_TYPES.get(file_ext)

        request_url = create_release_upload_response.upload_domain
        request_url += (
            f"/upload/set_metadata/{create_release_upload_response.package_asset_id}?"
        )

        parameters = {"file_name": file_name, "file_size": file_size}

        if mime_type:
            parameters["content_type"] = mime_type

        request_url += urllib.parse.urlencode(parameters)
        request_url += "&token=" + create_release_upload_response.url_encoded_token

        for attempt in range(3):
            self.log.debug(f"Attempting post {attempt}/3 in set_upload_metadata")
            try:
                response = self.post(request_url, data={})
                if response.ok:
                    return deserialize.deserialize(
                        SetUploadMetadataResponse, response.json()
                    )
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to post in set_upload_metadata: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return None

    def _upload_chunk(
        self,
        *,
        chunk_number: int,
        chunk: bytearray,
        create_release_upload_response: CreateReleaseUploadResponse,
    ) -> Optional[SetUploadMetadataResponse]:
        """Set the metadata for a binary upload

        :param CreateReleaseUploadResponse create_release_upload_response: The response to a `get_upload_url` call
        :param binary_path: The path to the binary to upload

        :returns: The upload response if we manage to get one, None otherwise
        """

        request_url = create_release_upload_response.upload_domain
        request_url += (
            f"/upload/upload_chunk/{create_release_upload_response.package_asset_id}?"
        )

        parameters = {"block_number": chunk_number}

        request_url += urllib.parse.urlencode(parameters)
        request_url += "&token=" + create_release_upload_response.url_encoded_token

        for attempt in range(3):
            self.log.debug(f"Attempting post {attempt}/3 in _upload_chunk")
            try:
                response = self.post_raw_data(url=request_url, data=chunk)
                if response.ok:
                    return deserialize.deserialize(ChunkUploadResponse, response.json())
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to post in _upload_chunk: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return None

    def _mark_upload_finished(
        self, *, create_release_upload_response: CreateReleaseUploadResponse
    ) -> Optional[UploadCompleteResponse]:
        """Mark the upload of a binary as finished

        :param CreateReleaseUploadResponse create_release_upload_response: The response to a `get_upload_url` call

        :returns: The upload complete response on success, None otherwise.
        """

        request_url = create_release_upload_response.upload_domain
        request_url += (
            f"/upload/finished/{create_release_upload_response.package_asset_id}?"
        )

        parameters = {"callback": ""}

        request_url += urllib.parse.urlencode(parameters)
        request_url += "&token=" + create_release_upload_response.url_encoded_token

        for attempt in range(3):
            self.log.debug(f"Attempting post {attempt}/3 in _mark_upload_finished")
            try:
                response = self.post_raw_data(request_url, data=None)
                if response.ok:
                    return deserialize.deserialize(
                        UploadCompleteResponse, response.json()
                    )
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to post in _mark_upload_finished: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return None

    def upload_binary(
        self,
        *,
        create_release_upload_response: CreateReleaseUploadResponse,
        binary_path: str,
    ) -> bool:
        """Upload a binary

        :param CreateReleaseUploadResponse create_release_upload_response: The response to a `get_upload_url` call
        :param binary_path: The path to the binary to upload

        :returns: True on success, False on failure
        """

        upload_metadata_response = self.set_upload_metadata(
            create_release_upload_response=create_release_upload_response,
            binary_path=binary_path,
        )

        if not upload_metadata_response:
            self.log.error("Failed to get upload metadata response")
            return False

        with open(binary_path, "rb") as binary_file:
            chunk_numbers = upload_metadata_response.chunk_list
            unhandled_chunks = []

            def direct_upload_chunk(chunk, chunk_number):
                nonlocal unhandled_chunks
                try:
                    response = self._upload_chunk(
                        chunk_number=chunk_number,
                        chunk=chunk,
                        create_release_upload_response=create_release_upload_response,
                    )
                    if response is None:
                        self.log.warn(
                            f"Failed to get response for uploading chunk {chunk_number}"
                        )
                        unhandled_chunks.append((chunk_number, 0, chunk))
                except Exception as ex:
                    self.log.warn(
                        f"Got an error response for uploading chunk {chunk_number} -> {ex}"
                    )
                    unhandled_chunks.append((chunk_number, 0, chunk))

            while len(chunk_numbers) > 0:
                chunk_number = chunk_numbers.pop(0)
                chunk = binary_file.read(upload_metadata_response.chunk_size)
                direct_upload_chunk(chunk, chunk_number)

            while len(unhandled_chunks) > 0:
                chunk_number, attempts, chunk = unhandled_chunks.pop(0)
                if attempts >= 3:
                    self.log.error(f"Failed to upload {len(unhandled_chunks)}")
                    return False
                direct_upload_chunk(chunk, chunk_number)

        self._mark_upload_finished(
            create_release_upload_response=create_release_upload_response
        )

        return True

    def commit_upload(
        self, *, owner_name: str, app_name: str, upload_id: str
    ) -> CommitUploadResponse:
        """Get the App Center release identifier for the app version (usually build number).

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str upload_id: The ID of the upload to commit

        :returns: The App Center release end response
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/uploads/releases/{upload_id}"

        data = {"upload_status": "uploadFinished"}

        for attempt in range(3):
            self.log.debug(f"Attempting patch {attempt}/3 in commit_upload")
            try:
                response = self.patch(request_url, data=data)
                if response.ok:
                    break
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to patch in commit_upload: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return deserialize.deserialize(CommitUploadResponse, response.json())

    def _wait_for_upload(
        self, *, owner_name: str, app_name: str, upload_id: str
    ) -> CommitUploadResponse:
        """Wait for an upload to finish processing

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str upload_id: The ID of the upload to wait for
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/uploads/releases/{upload_id}"

        def wait():
            self.log.info("Sleeping for 2 seconds before checking upload status again.")
            time.sleep(2)  # Same as the app center CLI

        while True:
            self.log.info("Checking upload status...")
            response = self.get(request_url)

            if not response.ok:
                wait()
                continue

            try:
                response_data = deserialize.deserialize(
                    CommitUploadResponse, response.json()
                )
            except Exception as ex:
                self.log.warning(f"Failed to get response data: {ex}")
                wait()
                continue

            if response_data.upload_status in ["uploadStarted", "uploadFinished"]:
                wait()
                continue

            if response_data.upload_status == "uploadCanceled":
                return response_data

            if response_data.upload_status == "readyToBePublished":
                return response_data

            if response_data.upload_status == "malwareDetected":
                raise Exception("Malware detected in uploaded binary")

            if response_data.upload_status == "error":
                raise Exception(
                    "An error occurred when waiting for binary: "
                    + response_data.get("error_details", "?")
                )

            raise Exception(f"Unexpected status: {response_data.upload_status}")

    def release(
        self,
        *,
        owner_name: str,
        app_name: str,
        release_id: int,
        group_id: str,
        mandatory_update: bool = False,
        notify_testers: bool = False,
    ) -> ReleaseDestinationResponse:
        """Release a build to a group.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param int release_id: The release ID of the app
        :param str group_id: The release ID of the group to release to
        :param bool mandatory_update: Set to True to make this a mandatory update
        :param bool notify_testers: Set to True to notify testers about this new build

        :returns: The App Center release identifier
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/releases/{release_id}/groups"

        data = {
            "id": group_id,
            "mandatory_update": mandatory_update,
            "notify_testers": notify_testers,
        }

        for attempt in range(3):
            self.log.debug(f"Attempting post {attempt}/3 in release")
            try:
                response = self.post(request_url, data=data)
                if response.ok:
                    break
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to post in release: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

        return deserialize.deserialize(ReleaseDestinationResponse, response.json())

    def update_release(
        self,
        *,
        owner_name: str,
        app_name: str,
        release_id: int,
        release_update_request: ReleaseUpdateRequest,
    ) -> None:
        """Update a release with new details

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param int release_id: The release ID of the app
        :param ReleaseUpdateRequest release_update_request: The release ID of the group to release to

        :returns: The App Center release identifier
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/releases/{release_id}"

        for attempt in range(3):
            self.log.debug(f"Attempting patch {attempt}/3 in update_release")
            try:
                response = self.patch(request_url, data=release_update_request.json())
                if response.ok:
                    break
            except Exception as ex:
                if attempt < 2:
                    self.log.warning(f"Failed to patch in update_release: {ex}")
                    self.log.warning("Will wait 10 seconds and try again")
                    time.sleep(10)
                else:
                    raise

    def upload_build(
        self,
        *,
        owner_name: str,
        app_name: str,
        binary_path: str,
        release_notes: str,
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        commit_message: Optional[str] = None,
    ) -> Optional[int]:
        """Get the App Center release identifier for the app version (usually build number).

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str binary_path: The path to the binary to upload
        :param str release_notes: The release notes for the release
        :param Optional[str] branch_name: The git branch that the build came from
        :param Optional[str] commit_hash: The hash of the commit that was just built
        :param Optional[str] commit_message: The message of the commit that was just built

        :raises FileNotFoundError: If the supplied binary is not found
        :raises Exception: If we don't get a release ID back after upload

        :returns: The release ID
        """

        if not os.path.exists(binary_path):
            raise FileNotFoundError(f"Could not find binary: {binary_path}")

        create_release_upload_response = self.get_upload_url(
            owner_name=owner_name, app_name=app_name
        )

        success = self.upload_binary(
            create_release_upload_response=create_release_upload_response,
            binary_path=binary_path,
        )

        if not success:
            raise Exception("Failed to upload binary")

        self.commit_upload(
            owner_name=owner_name,
            app_name=app_name,
            upload_id=create_release_upload_response.identifier,
        )

        upload_end_response = self._wait_for_upload(
            owner_name=owner_name,
            app_name=app_name,
            upload_id=create_release_upload_response.identifier,
        )

        if upload_end_response.release_distinct_id is None:
            raise Exception("No release ID was supplied in the upload end response")

        build_info = BuildInfo(
            branch_name=branch_name,
            commit_hash=commit_hash,
            commit_message=commit_message,
        )
        update_request = ReleaseUpdateRequest(
            release_notes=release_notes, build=build_info
        )

        self.update_release(
            owner_name=owner_name,
            app_name=app_name,
            release_id=upload_end_response.release_distinct_id,
            release_update_request=update_request,
        )

        return upload_end_response.release_distinct_id

    def upload_and_release(
        self,
        *,
        owner_name: str,
        app_name: str,
        binary_path: str,
        group_id: str,
        release_notes: str,
        notify_testers: Optional[bool] = None,
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        commit_message: Optional[str] = None,
    ) -> ReleaseDetailsResponse:
        """Get the App Center release identifier for the app version (usually build number).

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str binary_path: The path to the binary to upload
        :param str group_id: The ID of the group to release to
        :param str release_notes: The release notes for the release
        :param Optional[bool] notify_testers: Set to True to notify testers about this build
        :param Optional[str] branch_name: The git branch that the build came from
        :param Optional[str] commit_hash: The hash of the commit that was just built
        :param Optional[str] commit_message: The message of the commit that was just built

        :raises FileNotFoundError: If the supplied binary is not found
        :raises Exception: If we don't get a release ID back after upload

        :returns: The release details
        """

        release_id = self.upload_build(
            owner_name=owner_name,
            app_name=app_name,
            binary_path=binary_path,
            release_notes=release_notes,
            branch_name=branch_name,
            commit_hash=commit_hash,
            commit_message=commit_message,
        )

        if release_id is None:
            raise Exception("Did not get release ID after upload")

        self.release(
            owner_name=owner_name,
            app_name=app_name,
            release_id=release_id,
            group_id=group_id,
            notify_testers=notify_testers if notify_testers else False,
        )

        return self.release_details(
            owner_name=owner_name, app_name=app_name, release_id=release_id
        )



"""Base definition for App Center clients."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import Any, BinaryIO, Callable, Dict, Optional, Tuple

import deserialize
import requests
from tenacity import (
    retry,
    retry_if_exception,
    retry_if_result,
    stop_after_attempt,
    wait_fixed,
)


from appcenter.constants import API_BASE_URL


ProgressCallback = Callable[[int, Optional[int]], None]


class AppCenterHTTPError:
    """Represents the response for a HTTP error."""

    code: str
    message: str

    def __str__(self) -> str:
        """Generate and return the string representation of the object.
        :return: A string representation of the object
        """
        return f"<code={self.code}, message={self.message}>"


class AppCenterHTTPException(Exception):
    """All App Center HTTP exceptions use this class."""

    response: requests.Response

    def __init__(self, response: requests.Response) -> None:
        """Create a new AppCenterHTTPException

        :param response: The response to the HTTP request
        """

        super().__init__()
        self.response = response

    def __str__(self) -> str:
        """Generate and return the string representation of the object.
        :return: A string representation of the object
        """
        return (
            f"<method={self.response.request.method}, "
            + f"url={self.response.request.url}, "
            + f"status_code={self.response.status_code}, "
            + f"text={self.response.text}>"
        )


class AppCenterDecodedHTTPException(AppCenterHTTPException):
    """An App Center HTTP exception where we managed to decode the response."""

    error: AppCenterHTTPError

    def __init__(self, response: requests.Response, error: AppCenterHTTPError) -> None:
        """Create a new AppCenterDecodedHTTPException.

        :param response: The response to the HTTP request
        :param error: The decoded error
        """

        super().__init__(response)
        self.error = error

    def __str__(self) -> str:
        """Generate and return the string representation of the object.
        :return: A string representation of the object
        """
        return (
            f"method={self.response.request.method}, "
            + f"url={self.response.request.url}, "
            + f"status_code={self.response.status_code}, "
            + f"error={self.error}"
        )


def create_exception(response: requests.Response) -> AppCenterHTTPException:
    """Create the exception for a response.

    :param request: The original request
    :param response: The response from App Center

    :returns: The constructed exception
    """
    try:
        response_json = response.json()
        return AppCenterDecodedHTTPException(
            response, deserialize.deserialize(AppCenterHTTPError, response_json)
        )
    except Exception:
        return AppCenterHTTPException(response)


def _is_connection_failure(exception: BaseException) -> bool:
    exception_checks = [
        "Operation timed out",
        "Connection aborted.",
        "bad handshake: ",
        "Failed to establish a new connection",
    ]

    for check in exception_checks:
        if check in str(exception):
            return True

    return False


class AppCenterDerivedClient:
    """Base definition for App Center clients.

    :param name: The name of the derived client
    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    log: logging.Logger
    token: str
    session: requests.Session

    def __init__(self, name: str, token: str, parent_logger: logging.Logger) -> None:
        self.log = parent_logger.getChild(name)
        self.token = token
        self.session = requests.Session()
        self.session.headers.update({"X-API-Token": self.token})

    def base_url(self, *, version: str = "0.1") -> str:
        """Generate the base URL for the API.

        :param str version: The API version to hit

        :returns: The base URL
        """
        return f"{API_BASE_URL}/v{version}"

    def generate_url(
        self, *, version: str = "0.1", owner_name: str, app_name: str
    ) -> str:
        """Generate a URL to use for querying the API.

        :param str version: The API version to hit
        :param str owner_name: The name of the owner of the app
        :param str app_name: The name of the app

        :returns: A generated URL base
        """

        url = self.base_url(version=version) + f"/apps/{owner_name}/{app_name}"
        self.log.debug(f"Generated URL: {url}")
        return url

    @retry(
        retry=(
            retry_if_exception(_is_connection_failure)
            | retry_if_result(
                lambda response: response.status_code == 202
            )  # For a GET we also need to retry on a 202
        ),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def get(self, url: str) -> requests.Response:
        """Perform a GET request to a url

        :param url: The URL to run the GET on

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """

        response = self.session.get(url)

        if response.status_code != 200:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def patch(self, url: str, *, data: Any) -> requests.Response:
        """Perform a PATCH request to a url

        :param url: The URL to run the POST on
        :param data: The JSON serializable data to send

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """
        response = self.session.patch(
            url, headers={"Content-Type": "application/json"}, json=data
        )

        if response.status_code < 200 or response.status_code >= 300:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def post(self, url: str, *, data: Any) -> requests.Response:
        """Perform a POST request to a url

        :param url: The URL to run the POST on
        :param data: The JSON serializable data to send

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """
        response = self.session.post(
            url, headers={"Content-Type": "application/json"}, json=data
        )

        if response.status_code < 200 or response.status_code >= 300:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def post_raw_data(self, url: str, data: Any) -> requests.Response:
        """Perform a POST request to a url

        :param url: The URL to run the POST on
        :param data: The JSON serializable data to send

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """
        response = self.session.post(url, headers={}, data=data)

        if response.status_code < 200 or response.status_code >= 300:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def post_files(
        self, url: str, *, files: Dict[str, Tuple[str, BinaryIO]]
    ) -> requests.Response:
        """Perform a POST request to a url, sending files

        :param url: The URL to run the POST on
        :param files: The dictionary of file data. Each key is a unique file. The values are a tuple
                      with the first property being the file name, the second being a stream of the
                      file contents.

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """

        response = self.session.post(url, files=files, timeout=20 * 60)

        if response.status_code < 200 or response.status_code >= 300:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def delete(self, url: str) -> requests.Response:
        """Perform a DELETE request to a url

        :param url: The URL to run the DELETE on

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """
        response = self.session.delete(url)

        if response.status_code < 200 or response.status_code >= 300:
            raise create_exception(response)

        return response

    @retry(
        retry=(retry_if_exception(_is_connection_failure)),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3),
    )
    def azure_blob_upload(
        self, url: str, *, file_stream: BinaryIO
    ) -> requests.Response:
        """Upload a file to an Azure Blob Storage URL

        :param url: The URL to upload to
        :param file_stream: The stream to the file contents

        :returns: The raw response

        :raises AppCenterHTTPException: If the request fails with a non 200 status code
        """

        response = requests.put(
            url,
            headers={
                "Content-Type": "application/octet-stream",
                "x-ms-blob-type": "BlockBlob",
            },
            data=file_stream.read(),
        )

        if response.status_code < 200 or response.status_code >= 300:
            self.log.debug("Azure URL: " + url)
            raise create_exception(response)

        return response



#!/usr/bin/env python3

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""App Center API wrapper."""

import json
import logging
import re
import time
from typing import Any, ClassVar, Dict, List, Optional

import requests

from appcenter.account import AppCenterAccountClient
from appcenter.analytics import AppCenterAnalyticsClient
from appcenter.crashes import AppCenterCrashesClient
from appcenter.tokens import AppCenterTokensClient
from appcenter.versions import AppCenterVersionsClient


class AppCenterClient:
    """Class responsible for getting data from App Center through REST calls.

    :param str access_token: The access token to use for authentication. Leave as None to use KeyVault
    """

    log: logging.Logger
    token: str

    account: AppCenterAccountClient
    analytics: AppCenterAnalyticsClient
    crashes: AppCenterCrashesClient
    tokens: AppCenterTokensClient
    versions: AppCenterVersionsClient

    def __init__(
        self, *, access_token: str, parent_logger: Optional[logging.Logger] = None
    ) -> None:
        """Initialize the AppCenterClient with the application id and the token."""

        if parent_logger is None:
            self.log = logging.getLogger("appcenter")
        else:
            self.log = parent_logger.getChild("appcenter")

        self.token = access_token
        self.account = AppCenterAccountClient(self.token, self.log)
        self.analytics = AppCenterAnalyticsClient(self.token, self.log)
        self.crashes = AppCenterCrashesClient(self.token, self.log)
        self.tokens = AppCenterTokensClient(self.token, self.log)
        self.versions = AppCenterVersionsClient(self.token, self.log)



"""Data type models"""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import datetime
import enum
from typing import Any, Dict, List, Optional

import deserialize


def iso8601parse(date_string: Optional[str]) -> Optional[datetime.datetime]:
    """Parse an ISO8601 date string into a datetime.

    :param date_string: The date string to parse

    :returns: The parsed datetime
    """
    if date_string is None:
        return None
    try:
        return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")


# pylint: disable=missing-docstring


class HandledErrorReasonFrame:
    class ProgrammingLanguage(enum.Enum):
        JAVASCRIPT = "JavaScript"
        C_SHARP = "CSharp"
        OBJECTIVE_C = "Objective-C"
        OBJECTIVE_CPP = "Objective-Cpp"
        CPP = "Cpp"
        C = "C"
        SWIFT = "Swift"
        JAVA = "Java"
        UNKNOWN = "Unknown"

    className: Optional[str]  # name of the class
    method: Optional[str]  # name of the method
    classMethod: Optional[bool]  # is a class method
    file: Optional[str]  # name of the file
    line: Optional[int]  # line number
    appCode: Optional[bool]  # this line isn't from any framework
    frameworkName: Optional[str]  # Name of the framework
    codeFormatted: Optional[str]  # Formatted frame string
    codeRaw: Optional[str]  # Unformatted Frame string
    methodParams: Optional[str]  # parameters of the frames method
    exceptionType: Optional[str]  # Exception type.
    osExceptionType: Optional[str]  # OS exception type. (aka. SIGNAL)
    language: Optional[ProgrammingLanguage]  # programming language of the frame


class ErrorGroupState(enum.Enum):
    OPEN = "Open"
    CLOSED = "Closed"
    IGNORED = "Ignored"


@deserialize.parser("firstOccurrence", iso8601parse)
@deserialize.parser("lastOccurrence", iso8601parse)
class ErrorGroupListItem:
    state: ErrorGroupState
    annotation: Optional[str]
    errorGroupId: str
    appVersion: str
    appBuild: Optional[str]
    count: int
    deviceCount: int
    firstOccurrence: datetime.datetime
    lastOccurrence: datetime.datetime
    exceptionType: Optional[str]
    exceptionMessage: Optional[str]
    exceptionClassName: Optional[str]
    exceptionClassMethod: Optional[bool]
    exceptionMethod: Optional[str]
    exceptionAppCode: Optional[bool]
    exceptionFile: Optional[str]
    exceptionLine: Optional[str]
    codeRaw: Optional[str]
    reasonFrames: Optional[List[HandledErrorReasonFrame]]


class ErrorGroups:
    nextLink: Optional[str]
    errorGroups: Optional[List[ErrorGroupListItem]]


@deserialize.parser("firstOccurrence", iso8601parse)
@deserialize.parser("lastOccurrence", iso8601parse)
class ErrorGroup:
    state: ErrorGroupState
    annotation: Optional[str]
    errorGroupId: str
    appVersion: str
    appBuild: Optional[str]
    count: int
    deviceCount: int
    firstOccurrence: datetime.datetime
    lastOccurrence: datetime.datetime
    exceptionType: Optional[str]
    exceptionMessage: Optional[str]
    exceptionClassName: Optional[str]
    exceptionClassMethod: Optional[bool]
    exceptionMethod: Optional[str]
    exceptionAppCode: Optional[bool]
    exceptionFile: Optional[str]
    exceptionLine: Optional[str]
    codeRaw: Optional[str]
    reasonFrames: Optional[List[HandledErrorReasonFrame]]


@deserialize.parser("timestamp", iso8601parse)
class HandledError:
    errorId: Optional[str]
    timestamp: Optional[datetime.datetime]
    deviceName: Optional[str]
    osVersion: Optional[str]
    osType: Optional[str]
    country: Optional[str]
    language: Optional[str]
    userId: Optional[str]


class HandledErrors:
    nextLink: Optional[str]
    errors: Optional[List[HandledError]]


@deserialize.parser("timestamp", iso8601parse)
@deserialize.parser("appLaunchTimestamp", iso8601parse)
class HandledErrorDetails:
    errorId: Optional[str]
    timestamp: Optional[datetime.datetime]
    deviceName: Optional[str]
    osVersion: Optional[str]
    osType: Optional[str]
    country: Optional[str]
    language: Optional[str]
    userId: Optional[str]
    name: Optional[str]
    reasonFrames: Optional[List[HandledErrorReasonFrame]]
    appLaunchTimestamp: Optional[datetime.datetime]
    carrierName: Optional[str]
    jailbreak: Optional[bool]
    properties: Optional[Dict[str, str]]


class ReleaseOrigin(enum.Enum):
    HOCKEY = "hockeyapp"
    APP_CENTER = "appcenter"


@deserialize.auto_snake()
class BuildInfo:
    branch_name: Optional[str]
    commit_hash: Optional[str]
    commit_message: Optional[str]

    def __init__(
        self,
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        commit_message: Optional[str] = None,
    ) -> None:
        self.branch_name = branch_name
        self.commit_hash = commit_hash
        self.commit_message = commit_message

    def json(self) -> Dict[str, Any]:
        result = {}

        if self.branch_name is not None:
            result["branch_name"] = self.branch_name

        if self.commit_hash is not None:
            result["commit_hash"] = self.commit_hash

        if self.commit_message is not None:
            result["commit_message"] = self.commit_message

        return result


class StoreType(enum.Enum):
    INTUNE = "intune"
    GOOGLE_PLAY = "googleplay"
    APPLE = "apple"
    NONE = "none"


class DestinationType(enum.Enum):
    GROUP = "group"
    STORE = "store"
    TESTER = "tester"


@deserialize.key("identifier", "id")
@deserialize.key("store_type", "type")
class Destination:
    identifier: str
    name: Optional[str]
    is_latest: Optional[bool]
    store_type: Optional[StoreType]
    publishing_status: Optional[str]
    destination_type: Optional[DestinationType]
    display_name: Optional[str]


@deserialize.key("identifier", "id")
@deserialize.parser("uploaded_at", iso8601parse)
class BasicReleaseDetailsResponse:
    identifier: int
    version: str
    origin: Optional[ReleaseOrigin]
    short_version: str
    enabled: bool
    uploaded_at: datetime.datetime
    destinations: Optional[List[Destination]]
    build: Optional[BuildInfo]


class ProvisioningProfileType(enum.Enum):
    ADHOC = "adhoc"
    ENTERPRISE = "enterprise"
    OTHER = "other"


@deserialize.key("identifier", "id")
@deserialize.parser("provisioning_profile_expiry_date", iso8601parse)
@deserialize.parser("uploaded_at", iso8601parse)
class ReleaseDetailsResponse:
    # ID identifying this unique release.
    identifier: int

    # The app's name (extracted from the uploaded release).
    app_name: str

    # The app's display name.
    app_display_name: str

    # The app's OS.
    app_os: Optional[str]

    # The release's version.
    version: str

    # The release's origin
    origin: Optional[ReleaseOrigin]

    # The release's short version.
    short_version: str

    # The release's release notes.
    release_notes: Optional[str]

    # The release's provisioning profile name.
    provisioning_profile_name: Optional[str]

    # The type of the provisioning profile for the requested app version.
    provisioning_profile_type: Optional[ProvisioningProfileType]

    # expiration date of provisioning profile in UTC format.
    provisioning_profile_expiry_date: Optional[datetime.datetime]

    # A flag that determines whether the release's provisioning profile is still extracted or not.
    is_provisioning_profile_syncing: Optional[bool]

    # The release's size in bytes.
    size: Optional[int]

    # The release's minimum required operating system.
    min_os: Optional[str]

    # The release's device family.
    device_family: Optional[str]

    # The release's minimum required Android API level.
    android_min_api_level: Optional[str]

    # The identifier of the apps bundle.
    bundle_identifier: Optional[str]

    # Hashes for the packages
    package_hashes: Optional[List[str]]

    # MD5 checksum of the release binary.
    fingerprint: Optional[str]

    # The uploaded time.
    uploaded_at: datetime.datetime

    # The URL that hosts the binary for this release.
    download_url: Optional[str]

    # A URL to the app's icon.
    app_icon_url: Optional[str]

    # The href required to install a release on a mobile device. On iOS devices will be prefixed
    # with itms-services://?action=download-manifest&url=
    install_url: Optional[str]

    destinations: Optional[List[Destination]]

    # In calls that allow passing udid in the query string, this value will hold the provisioning
    # status of that UDID in this release. Will be ignored for non-iOS platforms.
    is_udid_provisioned: Optional[bool]

    # In calls that allow passing udid in the query string, this value determines if a release can
    # be re-signed. When true, after a re-sign, the tester will be able to install the release from
    # his registered devices. Will not be returned for non-iOS platforms.
    can_resign: Optional[bool]

    build: Optional[BuildInfo]

    # This value determines the whether a release currently is enabled or disabled.
    enabled: bool

    # Status of the release.
    status: Optional[str]


class ReleaseWithDistributionGroup:
    release: int  # The release ID
    distribution_group: str  # The distribution group ID

    def __init__(self, release: int, distribution_group: str) -> None:
        self.release = release
        self.distribution_group = distribution_group


class ReleaseCount:
    release_id: str
    distribution_group: Optional[str]
    unique_count: int
    total_count: int


class ReleaseCounts:
    total: Optional[int]
    counts: List[ReleaseCount]


@deserialize.key("identifier", "id")
class SetUploadMetadataResponse:
    identifier: str
    error: bool
    chunk_size: int
    resume_restart: bool
    chunk_list: List[int]
    blob_partitions: int
    status_code: str


class ChunkUploadResponse:
    error: bool
    chunk_num: int
    error_code: str


@deserialize.key("identifier", "id")
class CreateReleaseUploadResponse:
    identifier: str
    upload_domain: str
    token: str
    url_encoded_token: str
    package_asset_id: str


@deserialize.key("identifier", "id")
class CommitUploadResponse:
    identifier: str
    upload_status: str
    release_distinct_id: Optional[int]


@deserialize.key("identifier", "id")
class UploadCompleteResponse:
    absolute_uri: str
    chunk_num: int
    error: bool
    error_code: Optional[str]
    location: str
    message: str
    raw_location: str
    state: str


@deserialize.key("identifier", "id")
class ReleaseDestinationResponse:
    identifier: str
    mandatory_update: bool
    provisioning_status_url: Optional[str]


@deserialize.key("identifier", "id")
class DestinationId:
    name: Optional[str]
    identifier: Optional[str]

    def __init__(
        self, *, name: Optional[str] = None, identifier: Optional[str] = None
    ) -> None:
        self.name = name
        self.identifier = identifier

    def json(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        if self.name is not None:
            result["name"] = self.name

        if self.identifier is not None:
            result["id"] = self.identifier

        return result


class ReleaseUpdateRequest:
    release_notes: Optional[str]
    mandatory_update: Optional[bool]
    destinations: Optional[List[DestinationId]]
    build: Optional[BuildInfo]
    notify_testers: Optional[bool]

    def __init__(
        self,
        *,
        release_notes: Optional[str] = None,
        mandatory_update: Optional[bool] = None,
        destinations: Optional[List[DestinationId]] = None,
        build: Optional[BuildInfo] = None,
        notify_testers: Optional[bool] = None,
    ) -> None:
        self.release_notes = release_notes
        self.mandatory_update = mandatory_update
        self.destinations = destinations
        self.build = build
        self.notify_testers = notify_testers

    def json(self) -> Dict[str, Any]:
        output: Dict[str, Any] = {}

        if self.release_notes is not None:
            output["release_notes"] = self.release_notes

        if self.mandatory_update is not None:
            output["mandatory_update"] = self.mandatory_update

        if self.destinations is not None:
            output["destinations"] = [
                destination.json() for destination in self.destinations
            ]

        if self.build is not None:
            output["build"] = self.build.json()

        if self.notify_testers is not None:
            output["notify_testers"] = self.notify_testers

        return output


class SymbolType(enum.Enum):
    APPLE = "Apple"
    JAVASCRIPT = "JavaScript"
    BREAKPAD = "Breakpad"
    PROGUARD = "AndroidProguard"
    UWP = "UWP"


@deserialize.parser("expiration_date", iso8601parse)
class SymbolUploadBeginResponse:
    symbol_upload_id: str
    upload_url: str
    expiration_date: datetime.datetime


class SymbolUploadStatus(enum.Enum):
    COMMITTED = "committed"
    ABORTED = "aborted"


class SymbolUploadEndRequest:
    status: SymbolUploadStatus


class Origin(enum.Enum):
    APP_CENTER = "appcenter"
    HOCKEY = "hockeyapp"
    CODEPUSH = "codepush"


class Permission(enum.Enum):
    MANAGER = "manager"
    DEVELOPER = "developer"
    VIEWER = "viewer"
    TESTER = "tester"


class Role(enum.Enum):
    ADMIN = "admin"
    COLLABORATOR = "collaborator"
    MEMBER = "member"


@deserialize.key("identifier", "id")
class User:
    # The unique ID of the user
    identifier: str

    # The avatar URL of the user
    avatar_url: Optional[str]

    # User is required to send an old password in order to change the password
    can_change_password: Optional[bool]

    # The full name of the user. Might for example be first and last name
    display_name: str

    # The email address of the user
    email: str

    # The unique name that is used to identify the user
    name: str

    # The permissions the user has for the app
    permissions: List[Permission]

    # The creation origin of this user
    origin: Origin


@deserialize.key("identifier", "id")
@deserialize.parser("created_at", iso8601parse)
class UserToken:
    # The unique ID of the token
    identifier: str

    # The user supplied description for the token
    description: str

    # The scope the token has
    scope: List[str]

    # The creation date
    created_at: datetime.datetime

    # The value of the token - Only set when creating a new tokern
    api_token: Optional[str]



"""App Center crashes API wrappers."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import datetime
import logging
import os
from typing import Any, Dict, Iterator, Optional
import urllib.parse

from azure.storage.blob import BlockBlobService
import deserialize

import appcenter.constants
from appcenter.derived_client import AppCenterDerivedClient, ProgressCallback
from appcenter.models import (
    ErrorGroup,
    ErrorGroups,
    ErrorGroupListItem,
    ErrorGroupState,
    HandledError,
    HandledErrors,
    HandledErrorDetails,
    SymbolType,
    SymbolUploadBeginResponse,
    SymbolUploadEndRequest,
    SymbolUploadStatus,
)


class AppCenterCrashesClient(AppCenterDerivedClient):
    """Wrapper around the App Center crashes APIs.

    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    def __init__(self, token: str, parent_logger: logging.Logger) -> None:
        super().__init__("crashes", token, parent_logger)

    def group_details(
        self, *, owner_name: str, app_name: str, error_group_id: str
    ) -> ErrorGroup:
        """Get the error group details.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str error_group_id: The ID of the error group to get the details for

        :returns: An ErrorGroup
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/errors/errorGroups/{error_group_id}"

        response = self.get(request_url)

        return deserialize.deserialize(ErrorGroup, response.json())

    def error_details(
        self, *, owner_name: str, app_name: str, error_group_id: str, error_id: str
    ) -> HandledErrorDetails:
        """Get the error details.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str error_group_id: The ID of the error group to get the details for
        :param str error_id: The ID of the error to get the details for

        :returns: A HandledErrorDetails
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/errors/errorGroups/{error_group_id}/errors/{error_id}"

        response = self.get(request_url)

        return deserialize.deserialize(HandledErrorDetails, response.json())

    def error_info_dictionary(
        self, *, owner_name: str, app_name: str, error_group_id: str, error_id: str
    ) -> HandledErrorDetails:
        """Get the full error info dictionary.

        This is the full details that App Center has on a crash. It is not
        parsed due to being different per platform.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str error_group_id: The ID of the error group to get the details for
        :param str error_id: The ID of the error to get the details for

        :returns: The raw full error info dictionary
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += (
            f"/errors/errorGroups/{error_group_id}/errors/{error_id}/download"
        )

        response = self.get(request_url)

        return deserialize.deserialize(Dict[str, Any], response.json())

    def set_annotation(
        self,
        *,
        owner_name: str,
        app_name: str,
        error_group_id: str,
        annotation: str,
        state: Optional[ErrorGroupState] = None,
    ) -> None:
        """Get the error group details.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str error_group_id: The ID of the error group to set the annotation on
        :param str annotation: The annotation text
        :param Optional[ErrorGroupState] state: The state to set the error group to

        The `state` parameter here does seem somewhat unusual, but it can't be
        helped unfortunately. The API requires that we set the state with the
        annotation. In order to work around this, the state can either be set
        explicitly in this call, or if it is set to `None` (the default), we'll
        read the existing state and use that.
        """

        if state is None:
            request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
            request_url += f"/errors/errorGroups/{error_group_id}"

            response = self.get(request_url)

            group = deserialize.deserialize(ErrorGroup, response.json())
            state = group.state

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/errors/errorGroups/{error_group_id}"

        self.patch(request_url, data={"state": state.value, "annotation": annotation})

    def get_error_groups(
        self,
        *,
        owner_name: str,
        app_name: str,
        start_time: datetime.datetime,
        end_time: Optional[datetime.datetime] = None,
        version: Optional[str] = None,
        app_build: Optional[str] = None,
        group_state: Optional[ErrorGroupState] = None,
        error_type: Optional[str] = None,
        order_by: Optional[str] = None,
        limit: int = 30,
    ) -> Iterator[ErrorGroupListItem]:
        """Get the error groups for an app.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param datetime.datetime start_time: The time to start getting error groups from
        :param Optional[datetime.datetime] end_time: The end time to get error groups from
        :param Optional[str] version: The version of the app to restrict the search to (if any)
        :param Optional[str] app_build: The build to restrict the search to (if any)
        :param Optional[ErrorGroupState] group_state: Set to filter to just this group state (open, closed, ignored)
        :param Optional[str] error_type: Set to filter to specific types of error (all, unhandledError, handledError)
        :param Optional[str] order_by: The order by parameter to pass in (this will be encoded for you)
        :param Optional[str] limit: The max number of results to return per request (should not go past 100)

        :returns: An iterator of ErrorGroupListItem
        """

        # pylint: disable=too-many-locals

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/errors/errorGroups?"

        parameters = {"start": start_time.replace(microsecond=0).isoformat()}

        if end_time:
            parameters["end"] = end_time.replace(microsecond=0).isoformat()

        if version:
            parameters["version"] = version

        if app_build:
            parameters["app_build"] = app_build

        if group_state:
            parameters["groupState"] = group_state.value

        if error_type:
            parameters["errorType"] = error_type

        if order_by:
            parameters["$orderby"] = order_by

        if limit:
            parameters["$top"] = str(limit)

        request_url += urllib.parse.urlencode(parameters)

        page = 0

        while True:
            page += 1

            self.log.info(f"Fetching page {page} of crash groups")

            response = self.get(request_url)

            error_groups = deserialize.deserialize(ErrorGroups, response.json())

            yield from error_groups.errorGroups

            if error_groups.nextLink is None:
                break

            request_url = appcenter.constants.API_BASE_URL + self._next_link_polished(
                error_groups.nextLink, owner_name, app_name
            )

        # pylint: disable=too-many-locals

    def errors_in_group(
        self,
        *,
        owner_name: str,
        app_name: str,
        error_group_id: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        model: Optional[str] = None,
        operating_system: Optional[str] = None,
    ) -> Iterator[HandledError]:
        """Get the errors in a group.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str error_group_id: The ID of the group to get the errors from
        :param Optional[datetime.datetime] start_time: The time to start getting error groups from
        :param Optional[datetime.datetime] end_time: The end time to get error groups from
        :param Optional[str] model: The device model to restrict the search to (if any)
        :param Optional[str] operating_system: The OS to restrict the search to (if any)

        :returns: An iterator of HandledError
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/errors/errorGroups/{error_group_id}/errors?"

        parameters: Dict[str, str] = {}

        if start_time:
            parameters["start"] = start_time.replace(microsecond=0).isoformat()

        if end_time:
            parameters["end"] = end_time.replace(microsecond=0).isoformat()

        if model:
            parameters["model"] = model

        if operating_system:
            parameters["os"] = operating_system

        request_url += urllib.parse.urlencode(parameters)

        page = 0

        while True:
            page += 1

            self.log.info(f"Fetching page {page} of crashes for group {error_group_id}")

            response = self.get(request_url)

            errors = deserialize.deserialize(HandledErrors, response.json())

            yield from errors.errors

            if errors.nextLink is None:
                break

            request_url = appcenter.constants.API_BASE_URL + self._next_link_polished(
                errors.nextLink, owner_name, app_name
            )

    def begin_symbol_upload(
        self,
        *,
        owner_name: str,
        app_name: str,
        symbols_name: str,
        symbol_type: SymbolType,
        build_number: Optional[str] = None,
        version: Optional[str] = None,
    ) -> SymbolUploadBeginResponse:
        """Upload debug symbols

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str symbols_path: The path to the symbols
        :param str symbol_type: The type of symbols being uploaded
        :param Optional[str] build_number: The build number (required for Android)
        :param Optional[str] version: The build version (required for Android)

        :raises ValueError: If the build number or version aren't specified and it's an Android upload

        :returns: SymbolUploadBeginResponse
        """

        if symbol_type == SymbolType.PROGUARD:
            if build_number is None:
                raise ValueError("The build number is required for Android")

            if version is None:
                raise ValueError("The version is required for Android")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/symbol_uploads"

        data = {"symbol_type": symbol_type.value, "file_name": symbols_name}

        if build_number:
            data["build"] = build_number

        if version:
            data["version"] = version

        response = self.post(request_url, data=data)

        return deserialize.deserialize(SymbolUploadBeginResponse, response.json())

    def commit_symbol_upload(
        self, *, owner_name: str, app_name: str, upload_id: str
    ) -> SymbolUploadEndRequest:
        """Commit a symbol upload operation

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str upload_id: The ID of the symbols upload to commit

        :returns: The App Center symbol upload end response
        """

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/symbol_uploads/{upload_id}"

        data = {"status": "committed"}

        response = self.patch(request_url, data=data)

        return deserialize.deserialize(SymbolUploadEndRequest, response.json())

    def upload_symbols(
        self,
        *,
        owner_name: str,
        app_name: str,
        symbols_path: str,
        symbols_name: Optional[str] = None,
        symbol_type: SymbolType,
        build_number: Optional[str] = None,
        version: Optional[str] = None,
        progress_callback: Optional[ProgressCallback] = None,
    ) -> None:
        """Upload debug symbols

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str symbols_path: The path to the symbols
        :param str symbols_name: The name to use for the symbols (defaults to file basename)
        :param str symbol_type: The type of symbols being uploaded
        :param Optional[str] build_number: The build number (required for Android)
        :param Optional[str] version: The build version (required for Android)
        :param Optional[ProgressCallback] progress_callback: The upload progress callback

        For the upload progress callback, this is a callable where the first
        parameter is the number of bytes uploaded, and the second parameter is
        the total number of bytes to upload (if known).

        :raises Exception: If something goes wrong
        :raises FileNotFoundError: If the symbols path doesn't have a file at the end
        """

        if not os.path.isfile(symbols_path):
            raise FileNotFoundError(f"There was no file at: {symbols_path}")

        if not symbols_name:
            symbols_name = os.path.basename(symbols_path)

        begin_upload_response = self.begin_symbol_upload(
            owner_name=owner_name,
            app_name=app_name,
            symbols_name=symbols_name,
            symbol_type=symbol_type,
            build_number=build_number,
            version=version,
        )

        with open(symbols_path, "rb") as symbols_file:
            url_components = urllib.parse.urlparse(begin_upload_response.upload_url)
            path = url_components.path[1:]
            container, blob_name = path.split("/")
            connection_string = (
                f"BlobEndpoint={url_components.scheme}://{url_components.netloc};"
            )
            connection_string += f"SharedAccessSignature={url_components.query}"
            service = BlockBlobService(connection_string=connection_string)
            service.create_blob_from_stream(
                container, blob_name, symbols_file, progress_callback=progress_callback
            )

        commit_response = self.commit_symbol_upload(
            owner_name=owner_name,
            app_name=app_name,
            upload_id=begin_upload_response.symbol_upload_id,
        )

        if commit_response.status != SymbolUploadStatus.COMMITTED:
            raise Exception("Failed to upload symbols")

    def _next_link_polished(
        self, next_link: str, owner_name: str, app_name: str
    ) -> str:
        """Polish nextLink string gotten from AppCenter service

        :param str next_link: The nextLink property from a service response when items are queried in batches
        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app

        :returns: A polished next link to use on next batch
        """

        _ = self

        if f"{owner_name}/{app_name}" not in next_link:
            # For some apps, AppCenter is returning invalid nextLinks without app name and owner, just a // instead.
            next_link = next_link.replace("//", f"/{owner_name}/{app_name}/", 1)

        # AppCenter is returning a nextLink with a /api on the URL which causes the request to fail.
        return next_link.replace("/api", "", 1)



"""App Center analytics API wrappers."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import List

import deserialize

from appcenter.derived_client import AppCenterDerivedClient
from appcenter.models import ReleaseWithDistributionGroup, ReleaseCounts


class AppCenterAnalyticsClient(AppCenterDerivedClient):
    """Wrapper around the App Center analytics APIs.

    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    def __init__(self, token: str, parent_logger: logging.Logger) -> None:
        super().__init__("analytics", token, parent_logger)

    def release_counts(
        self,
        *,
        owner_name: str,
        app_name: str,
        releases: List[ReleaseWithDistributionGroup],
    ) -> ReleaseCounts:
        """Get the release counts for an app

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param List[ReleaseWithDistributionGroup] releases: The list of releases to get the counts for

        :returns: The release counts
        """

        self.log.info(f"Getting release counts for: {owner_name}/{app_name}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/analytics/distribution/release_counts"

        data = []
        for release in releases:
            data.append(
                {
                    "release": release.release,
                    "distribution_group": release.distribution_group,
                }
            )

        response = self.post(request_url, data={"releases": data})

        return deserialize.deserialize(ReleaseCounts, response.json())



"""App Center account API wrappers."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import List, Optional
import urllib.parse

import deserialize

from appcenter.derived_client import AppCenterDerivedClient
from appcenter.models import User, Permission, Role


class AppCenterAccountClient(AppCenterDerivedClient):
    """Wrapper around the App Center account APIs.

    :param token: The authentication token
    :param parent_logger: The parent logger that we will use for our own logging
    """

    def __init__(self, token: str, parent_logger: logging.Logger) -> None:
        super().__init__("account", token, parent_logger)

    def users(self, *, owner_name: str, app_name: str) -> List[User]:
        """Get the users for an app

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app

        :returns: The list of users
        """

        self.log.info(f"Getting user for: {owner_name}/{app_name}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/users"

        response = self.get(request_url)

        return deserialize.deserialize(List[User], response.json())

    def add_collaborator(
        self,
        *,
        owner_name: str,
        app_name: str,
        user_email: str,
        role: Optional[Role] = None,
    ) -> None:
        """Add a user as a collaborator to an app.

        If they are a new collaborator, they will be invited. If that is the
        case, the role must be specified.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str user_email: The email of the user
        :param Optional[Role] role: The role the user should have (this is required for new users)

        :returns: The list of users
        """

        self.log.info(
            f"Adding user {user_email} as collaborator on: {owner_name}/{app_name}"
        )

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += "/invitations"

        data = {"user_email": user_email}

        if role:
            data["role"] = role.value

        self.post(request_url, data=data)

    def delete_collaborator(
        self, *, owner_name: str, app_name: str, user_email: str
    ) -> None:
        """Remove a user as a collaborator from an app.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str user_email: The email of the user to remove
        """

        self.log.info(f"Deleting user {user_email} from: {owner_name}/{app_name}")

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/users/{urllib.parse.quote(user_email)}"

        self.delete(request_url)

    def set_collaborator_permissions(
        self, *, owner_name: str, app_name: str, user_email: str, permission: Permission
    ) -> None:
        """Set a users collaborator permissions on an app.

        :param str owner_name: The name of the app account owner
        :param str app_name: The name of the app
        :param str user_email: The email of the user
        :param Permission permission: The permission level to grant

        :returns: The list of users
        """

        self.log.info(
            f"Setting user {user_email} as collaborator with permission "
            + f"{permission.value} on: {owner_name}/{app_name}"
        )

        request_url = self.generate_url(owner_name=owner_name, app_name=app_name)
        request_url += f"/users/{urllib.parse.quote(user_email)}"

        self.patch(request_url, data={"permissions": [permission.value]})
